#Author: Joseph Haraldson
#all rights reserved, copyright 2018


levenberg_marquardt_implicit_merit := proc(residuals,vars,x_init,tol:=1e-16,max_iter := 100000,lm_param:= 1,lm_amp_factor:=1e-1,lm_damp_factor:=1e1)
      local i,
	    grad,
	    J,
	    JTJ,
	    rhs_k, 
	    lhs_k,
	    current_pt,
	    x_next,
	    I,
	    x_cur,
	    mu_k
	    ;
	    
	    J := VectorCalculus:-Jacobian(residuals,vars);
	    JTJ := LinearAlgebra:-Transpose(J).J;
	    Ident:=Matrix(LinearAlgebra:-Dimension(JTJ));
	    for j from 1 to LinearAlgebra:-RowDimension(Ident) do 
		Ident[j,j]:=1:
	    od:
	    
	  
	  
	    
	    x_cur := subs(x_init, Vector(vars));
	    
	    mu_k := lm_param;
	    for i from 1 to max_iter do
		
		current_pt := subs(Equate(Vector(vars),x_cur));
		
		
		
		
		rhs_k := subs(current_pt,JTJ + mu_k * Ident);
		
		lhs_k := subs(current_pt,-1*LinearAlgebra:-Transpose(J).residuals);
		
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		x_next_candidate := Equate(Vector(vars),x_cur + dx);
		
		if norm(subs(x_next_candidate, residuals),2) <= norm(subs(current_pt,residuals),2)  then 
		    mu_k := max(mu_k*(lm_amp_factor), 10^(-.5*Digits)); 
		   
		   
		    
		    
		    
		    if norm(dx,2)<  norm(subs(x_next_candidate,residuals),2) and  norm(subs(x_next_candidate,residuals),2) <= 1.5 * norm(dx,2)   then 
			mu_k := norm(subs(x_next_candidate,residuals),2);
		    fi:
		    
		    x_next := x_cur + dx;
		    printf("k = %d, mu=%e, ||x_k - x_k+1|| = %e, ||dx|| = %e, ||f(x_k)||=%e\n",i,mu_k,norm(x_cur - x_next,2), norm(dx,2), norm(subs(current_pt,residuals),2));
		
		    if norm(x_cur-x_next,2) < tol or norm(subs(current_pt,residuals),2) < tol then break fi:
		
		    x_cur := x_next;
		
		
		else 
		    mu_k := mu_k * (lm_damp_factor);  
		fi:
		
		
		
	    od:
	    
	    current_pt := subs(Equate(Vector(vars),x_cur));
	    return [norm(subs(Equate(vars,x_cur),residuals),2) , Equate(vars,x_cur)];
	    
end proc:
levenberg_marquardt := proc(obj,residuals,vars,x_init,tol:=1e-16,max_iter := 100000,lm_param:= 1,lm_amp_factor:=1e-1,lm_damp_factor:=1e1)
      local i,
	    grad,
	    J,
	    JTJ,
	    rhs_k, 
	    lhs_k,
	    current_pt,
	    x_next,
	    I,
	    x_cur,
	    mu_k
	    ;
	    
	    J := VectorCalculus:-Jacobian(residuals,vars);
	    JTJ := LinearAlgebra:-Transpose(J).J;
	    Ident:=Matrix(LinearAlgebra:-Dimension(JTJ));
	    for j from 1 to LinearAlgebra:-RowDimension(Ident) do 
		Ident[j,j]:=1:
	    od:
	    
	  
	  
	    
	    x_cur := subs(x_init, Vector(vars));
	    
	    mu_k := lm_param;
	    for i from 1 to max_iter do
		
		current_pt := subs(Equate(Vector(vars),x_cur));
		
		
		
		
		rhs_k := subs(current_pt,JTJ + mu_k * Ident);
		
		lhs_k := subs(current_pt,-1*LinearAlgebra:-Transpose(J).residuals);
		
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		x_next_candidate := Equate(Vector(vars),x_cur + dx);
		
		if subs(x_next_candidate, obj) <= subs(current_pt,obj)  then 
		    mu_k := max(mu_k*(lm_amp_factor), 10^(-.5*Digits)); 
		   
		   
		    
		    
		    
		    if norm(dx,2)<  subs(x_next_candidate,obj) and  subs(x_next_candidate,obj) <= 1.5 * norm(dx,2)   then 
			mu_k := norm(subs(x_next_candidate,residuals),2);
		    fi:
		    
		    x_next := x_cur + dx;
		    printf("k = %d, mu=%e, ||x_k - x_k+1|| = %e, ||dx|| = %e, ||f(x_k)||=%e\n",i,mu_k,norm(x_cur - x_next,2), norm(dx,2), sqrt(subs(current_pt,obj)));
		
		    if norm(x_cur-x_next,2) < tol or sqrt(subs(current_pt,obj)) < tol then break fi:
		
		    x_cur := x_next;
		
		
		else 
		    mu_k := mu_k * (lm_damp_factor);  
		fi:
		
		
		
	    od:
	    
	    current_pt := subs(Equate(Vector(vars),x_cur));
	    return [sqrt(subs(Equate(vars,x_cur),obj)) , Equate(vars,x_cur)];
	    
end proc:
initial_guess_lagrangian := proc(Gx,X_init,Lambda)
	local lagrangian_system,
	      my_eqns,
	      cf_mat,
	      cf_vec,
	      ineq_cons,
	      ls_soln,
	      Lambda_init
	      ; 
    
	lagrangian_system :=  convert(subs(X_init,Gx),Vector);
	my_eqns := Equate(lagrangian_system, Vector(ArrayNumElems(lagrangian_system)));
	cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix(my_eqns,convert(Lambda,list));
	
	
	ls_soln := Optimization:-LSSolve([cf_vec,cf_mat])[2];
	Lambda_init := Equate(convert(Lambda,Vector), ls_soln);
	
	return Lambda_init;
    
end proc:
