#Author: Joseph Haraldson
#all rights reserved, copyright 2018


generalized_sylvester_matrix :=proc(L,x:='t')
    local N, 
	  d, 
	  d_alt, 
	  b_degrees, 
	  Syl, 
	  Syl_eqns, 
	  b, 
	  b_vars 
	  ;
	  
	  N:=nops(L); 
	  d := max(seq(degree(L[i],x),i=1..nops(L))); 
	  d_alt := max( {-1,seq(degree(L[i],x),i=1..nops(L))} minus {d} ); 
        
         if d_alt <0 then 
                d_alt := d
         fi:
         
	 
	 
	  b_degrees := Vector(nops(L)); 
	  for i from 1 to nops(L) do 
	      if degree(L[i],x) <> d then 
		  b_degrees[i] := d-1;
	      
	      else 
		  b_degrees[i]:=d_alt-1;
	      fi:
	  od:
	  
	  b:= `<,>`(seq(add(x^j*'bb'[i,j],j=0..b_degrees[i]),i=1..nops(L))); 
	  b_vars := [seq(seq('bb'[i,j],j=0..b_degrees[i]),i=1..nops(L))]; 
	  
	  Syl_eqns := collect(Vector[row](L) . b,x); 
	  
	  Syl_eqns:= PolynomialTools:-CoefficientVector(Syl_eqns,x);
	  Syl_eqns := Equate(Syl_eqns,Vector(ArrayNumElems(Syl_eqns)));
	  
	  Syl := LinearAlgebra:-GenerateMatrix(Syl_eqns, b_vars)[1];
	  
	  return Syl, [b,b_vars];
	  
end proc:
generate_generalized_sylvester_entries := proc(A,num_SNF:=2)
      local n,
      L,
      minors, 
      S,
      r,c 
      ;
      
      n := LinearAlgebra:-RowDimension(A); 
      L := []; 
      
      for i from 0 to num_SNF-1 do 
	    S:= combinat[choose](n,n-i);
	    L:=[op(L),seq(seq(LinearAlgebra:-Determinant(LinearAlgebra:-SubMatrix(A, r, c)), r = S), c = S)];
      od:
      
   
      
      return L;
end proc:
matrix_degree := proc( A,t)
    local i,
	  d;
	
    d := degree(A(1),t);  
    for i from 2 to ArrayNumElems(A) do 
	d := max(degree(A(i),t),d);
    od:
    return d;
end proc:
generate_perturbation:=proc(A,t,method:="support",symbol:='a')
	if method = "support" then 
	    return generate_perturbation_support(A,t,symbol);
	elif method = "degree" then 
	    return generate_perturbation_mat_degree(A,t,symbol);
	elif method = "entry" then 
	     return generate_perturbation_entry_degree(A,t,symbol);
	fi:
	
	return FAIL,FAIL;
end proc:
generate_perturbation_entry_degree := proc(A,t,symbol:='a')
      local i,
	    M,
	    j,
	    a_vars
	    ;
	    
	    M:= Matrix(LinearAlgebra:-Dimension(A));
	    a_vars := []:
	    
	    for i from 1 to LinearAlgebra:-RowDimension(A) do
		for j from 1 to LinearAlgebra:-ColumnDimension(A) do 
		    M[i,j] := add(symbol[i,j,k]*t^k, k=0..degree(A[i,j],t));
		    a_vars := [op(a_vars), seq(symbol[i,j,k], k=0..degree(A[i,j],t))];
		od:
	    od:
	    return M,a_vars;
end proc:
generate_perturbation_mat_degree:= proc(A,t, symbol:='a')
      local i,
	    M,
	    j,
	    d,
	    a_vars
	    ;
	    
	    M:= Matrix(LinearAlgebra:-Dimension(A));
	    a_vars := []:
	    d:= matrix_degree(A,t);
	    
	    for i from 1 to LinearAlgebra:-RowDimension(A) do
		for j from 1 to LinearAlgebra:-ColumnDimension(A) do 
		    M[i,j] := add(symbol[i,j,k]*t^k, k=0..d);
		    a_vars := [op(a_vars), seq(symbol[i,j,k], k=0..d)];
		od:
	    od:
	    return M,a_vars;
end proc:
generate_perturbation_support := proc(A,t,symbol:='a')
	local i,
	      M,
	      j,
	      m,
	      n,
	      k,
	      a_vars,
	      cf_vec;
	      ;
	      
	      
	      n,m := LinearAlgebra:-Dimension(A);
	      M:= Matrix(n,m);
	      a_vars := []:
	      
	      for i from 1 to n do 
		  for j from 1 to m do 
			cf_vec := PolynomialTools:-CoefficientVector(A[i,j],t);
			
			a_vars := [op(a_vars), seq(symbol[i,j,k], k=0..degree(A[i,j],t))];
			M[i,j] := add(symbol[i,j,k]*t^k,k=0..degree(A[i,j],t));
			
			
			for k from 1 to ArrayNumElems(cf_vec) do
			    if(cf_vec[k]=0) then
				a_vars := convert(convert(a_vars,set) minus convert(symbol[i,j,k-1],set),list);
				M[i,j]:=subs(symbol[i,j,k-1]=0,M[i,j]);
			    fi:
			od:
			
		  od:
	      od:
	      
	      
	      return M, a_vars;
end proc:
poly_norm_opt :=proc(A,x:=t,p:=2)
	local nm;
	local i;
	nm := add(norm_opt(A(i),x,2)^2, i=1..ArrayNumElems(A));
	return sqrt(nm);
end proc:
norm_opt:=proc(f,x,p:=2)
	local v;
	local nm;
	v:=PolynomialTools:-CoefficientVector(f,x);
	nm := add(v[i]^p,i=1..ArrayNumElems(v));
	return (nm)^(1/p);
end proc:
linearize_matrix_poly_eqn := proc(Ab,x,vars, clean_eqns := true)
    local i,
	  my_eqns
	  ;
	  
	  my_eqns := [];
	  
	  for i from 1 to ArrayNumElems(Ab) do 
	      my_eqns := [op(my_eqns),op(Equate(PolynomialTools:-CoefficientVector(Ab(i),x), Vector(degree(Ab(i),x)+1)))];
	  od:
	  
	  
	  if clean_eqns = true then 
	      my_eqns := convert(convert(my_eqns,set) minus{ 0=0},list);
	  fi:
	  return my_eqns;
end proc:
clean_up := proc (A, t, tol:=1e-4)
    local i,
	  j,
	  cf_vec
	  ;
	  for i from 1 to ArrayNumElems(A) do 
	      cf_vec:= PolynomialTools:-CoefficientVector(A(i),t);
	      for j from 1 to ArrayNumElems(cf_vec) do 
		  if abs(cf_vec(j)) < tol then 
		  cf_vec(j) := 0;
		  fi:
	      od:
	      A(i):= PolynomialTools:-FromCoefficientVector(cf_vec,t);
	  od:
	  
	  return A;
	  
end proc:
random_pt := proc(L)
    local i, M := []:
    
    for i from 1 to nops(L) do 
	M:= [op(M), L[i] = RandomTools:-MersenneTwister:-GenerateFloat()];
    od:
    return M;
end proc:
combined_coefficient_vector := proc (L, x)
      local cf_vec;
      
      cf_vec := [];
      
      cf_vec := [ seq(op(convert(PolynomialTools:-CoefficientVector(L[i],x),list)),i=1..nops(L))];
      
      return Vector(cf_vec);
end proc:
poly_least_squares_division := proc(residuals, fstar_vars)
      local  A,b;
      Optimization:-LSSolve(residuals,fstar_vars);
end proc:
generate_initial_agcd := proc(f,x:=t)
      
      local h_vars,
	    fstar,
	    Syl,
	    bez, 
	    bezout_system,
	    bez_vars,
	    residuals_1, 
	    residuals_2, 
	    h_vec, 
	    soln_1 ,
	    h,
	    fstar_init,
	    d,
	    fstar_vars,
	    fs_residuals,
	    i,
	    h_init,
	    soln_2,
	    cf_mat,
	    cf_vec,
	    vars,
	    my_soln,
	    default_fstar_pt
	    ;
      
	    h_vars := [hh[0],hh[1],hh[2]];
	    h := hh[0]  + hh[1]*x + hh[2]*x^2;
	   
	    
	    Syl,bez := generalized_sylvester_matrix(f,x);
	    
	    bez_vars := bez[2];
	    bez:= bez[1]; 
	    
	    bezout_system := PolynomialTools:-CoefficientVector(add(f[i] * bez[i], i=1..nops(f))-h,x);
	    
	    bezout_system := Equate(bezout_system, Vector(ArrayNumElems(bezout_system)));
	   
	   
	    residuals_1 :=  convert(subs( [hh[1]=1,hh[2]=0], bezout_system),list); 
	    residuals_2 :=  convert(subs( hh[2]=1, bezout_system),list);
	   
	   
	    
	    vars := [op(bez_vars), op(h_vars)];
	   
	    cf_mat, cf_vec := LinearAlgebra:-GenerateMatrix(residuals_1,vars);
	   
	    soln_1 := Optimization:-LSSolve([cf_vec,cf_mat]);
	    
	    cf_mat, cf_vec := LinearAlgebra:-GenerateMatrix(residuals_2,vars);
	    
	    soln_2 := Optimization:-LSSolve([cf_vec,cf_mat]);
	    
	    my_soln := soln_1;
	    
	    
	    if my_soln[1] >  soln_2[1] then
		  my_soln := soln_2;
		  h := subs(hh[2] =1, h);
	    else
		  h:= subs([hh[1]=1,hh[2]=0],h);
	    fi:
	   
	    h_init := subs(Equate( Vector(vars), my_soln[2]),h);
	    
	    
	
	    
	    d := [(seq( max(degree(f[i],x),1), i=1..nops(f)))]:
	    fstar := [seq( add(fs[i,j] * x^j, j=0..d[i]),i=1..nops(d))]:
	    fstar_vars := [seq( seq(fs[i,j] , j=0..d[i]),i=1..nops(d))]:
  
	    
	    fs_residuals := combined_coefficient_vector(convert(collect(Vector(f)-  Vector(zip(`*`,fstar,h)),x),list),x);
	   
	    fs_residuals := subs(Equate( Vector(h_vars), Vector(PolynomialTools:-CoefficientVector(h_init,x), nops(h_vars))) ,fs_residuals);
	    cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix( convert(fs_residuals,list),fstar_vars);
	    fstar_init := Optimization:-LSSolve([cf_vec,cf_mat])[2]:
	    fstar_init := Equate(Vector(fstar_vars),fstar_init);
	    
	    
	    
	    return fstar_init, Equate( Vector(h_vars),Vector(PolynomialTools:-CoefficientVector(h_init,x),nops(h_vars)));
	
end proc:
generate_initial_agcd_degree := proc(f,d,x:=t)
      
      local h_vars,
	    fstar,
	    Syl,
	    bez, 
	    bezout_system,
	    bez_vars,
	    residuals_1, 
	    residuals_2, 
	    h_vec, 
	    soln_1 ,
	    h,
	    fstar_init,
	    fstar_vars,
	    fs_residuals,
	    i,
	    h_init,
	    cf_mat,
	    cf_vec,
	    vars,
	    my_soln,
	    default_fstar_pt
	    ;
      
	  
	  
	    
	    h_vars := [seq(hh[i],i=0..d-1)];
	    h := add(hh[i]*x^i, i=0..d-1) + x^d; 
	   
	    
	    Syl,bez := generalized_sylvester_matrix(f,x);
	    
	    bez_vars := bez[2];
	    bez:= bez[1]; 
	    
	    bezout_system := PolynomialTools:-CoefficientVector(add(f[i] * bez[i], i=1..nops(f))-h,x);
	    
	    bezout_system := Equate(bezout_system, Vector(ArrayNumElems(bezout_system)));
	   
	   
	    residuals_1 :=  convert(bezout_system,list); 
	   
	   
	    
	    vars := [op(bez_vars), op(h_vars)];
	   
	    cf_mat, cf_vec := LinearAlgebra:-GenerateMatrix(residuals_1,vars);
	   
	    soln_1 := Optimization:-LSSolve([cf_vec,cf_mat]);
	    
	    if DEBUG = true  then
		printf("The generalized Sylvester matrix has RowDim: %d and Rank %d\n",LinearAlgebra:-RowDimension(Syl), LinearAlgebra:-Rank(Syl));
	    fi:
	    
	    return subs(Equate(vars,soln_1[2]) ,h );
end proc:
poly_least_squares_division_implicit := proc(f,h,x:=t)
      local fstar,
	    fs_vars,
	    cf_vec,cf_mat,
	    residuals,
	    ls_soln
	    ;
	    
	    fstar := [ seq(  add(fs[i,j]*t^j,j=0..degree(f[i],x)-degree(h,x)) ,i=1..nops(f)) ];
	    fstar_vars := convert(`<,>`(seq( PolynomialTools:-CoefficientVector(fstar[i],x)  ,i=1..nops(f))),Vector);
	    
	    residuals :=  convert(`<,>`(seq( PolynomialTools:-CoefficientVector(f[i]-fstar[i]*h,x)  ,i=1..nops(f))),list);
	    
	    
	    cf_mat, cf_vec := LinearAlgebra:-GenerateMatrix(residuals,convert(fstar_vars,list));
	    
	    
	    ls_soln := Optimization:-LSSolve([cf_vec,cf_mat]);
	   
	    return [ls_soln[1], subs(Equate(fstar_vars,ls_soln[2]),fstar)];
end proc:
force_columns_prim := proc(A,x:=t,tol:=1e-10)
        local   S,
                K_prim,
                agcd, agcd_degree,
                i,j,
                my_rank,
                SS,
                current_col ; 
                ;
                
                K_prim := Matrix(0,0):
                for i from 1 to LinearAlgebra:-ColumnDimension(A) do 
                        current_col := convert(A[1..LinearAlgebra:-RowDimension(A), i],list);
                        S:= generalized_sylvester_matrix( current_col,x)[1];
                        
                        my_rank := LinearAlgebra:-RowDimension(S):
                        SS:=LinearAlgebra:-SingularValues(evalf(S));
                        for j from 1 to ArrayNumElems(SS) do
                                if abs(SS[j]) > tol then
                                    my_rank := j;
                                fi:
                        od:
               
                       
                       agcd_degree:=LinearAlgebra:-RowDimension(S)-my_rank;
                       
                        agcd:=generate_initial_agcd_degree(current_col,agcd_degree,x);
                        
                        if ArrayNumElems(K_prim) = 0 then 
                        K_prim := convert(poly_least_squares_division_implicit(current_col,agcd,x)[2],Vector);
                        else
                            K_prim := `<|>`(K_prim,convert(poly_least_squares_division_implicit(current_col,agcd,x)[2],Vector));
                        fi:
                od:
            return K_prim;
end proc:
