#Author: Joseph Haraldson
#all rights reserved, copyright 2018


nearest_snf := proc(A,t,DeltaA,DeltaA_init)
      local i,
	    j,
	    Lambda,
	    Lambda_init,
	    adj,
	    adj_vec,
	    fstar,
	    h,
	    X,
	    Z,
	    X_init,
	    Z_init
	    ;
	    
	    ;
end proc:
symbolic_numeric_lm := proc(my_args,tol:=1e-16,max_iter := 100000,lm_param:= 1,lm_amp_factor:=1e-1,lm_damp_factor:=1e1)
      local i,
	    grad,
	    J,
	    JTJ,
	    rhs_k, 
	    lhs_k,
	    current_pt,
	    x_next,
	    Imat,
	    residuals,
	    obj,
	    t_start,t_end,t_iter,
	    j,
	    dx,
	    x_next_candidate,
	    x_cur,
	    A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init,Z,
	    Z_next_candidate,
	    mu_k,
	    safety_factor 
	    ;
	    t_start := time():
	    
	    A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init,Z := op(my_args);
	    
	    safety_factor:= 10^(-1*Digits);
	    
	    
	   
	    J:= fast_symbolic_hessian_full(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init);
	    JTJ := LinearAlgebra:-Transpose(J).J;
	    printf("Computed Initial Numerical Hessian...\n");
	    
	    residuals := fast_symbolic_gradient(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init);
	    obj := norm(residuals,2);
	    
	    
	    printf("Computed numerical gradient...\n");
	    
	    
	    
	    Imat:=Matrix(LinearAlgebra:-Dimension(JTJ));
	    for j from 1 to LinearAlgebra:-RowDimension(Imat) do 
		Imat[j,j]:=1:
	    od:
	    
	  
	    
	    x_cur := convert(subs(Z_init,Z),Vector);
	    mu_k := lm_param;
	    
	    printf("Beginning iterations. Setuptime is %f \n", time()-t_start);
	  
	    for i from 1 to max_iter do
		rhs_k := JTJ + mu_k * Imat;
		lhs_k := -1*LinearAlgebra:-Transpose(J).residuals;
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		
		x_next_candidate := x_cur + dx;
		
		
		Z_next_candidate := Equate(Z,x_next_candidate): 
		residuals_next :=  fast_symbolic_gradient(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_next_candidate);
		obj_next := norm(residuals_next,2);
		
		
	
		if obj_next <= obj  then 
		
		    mu_k :=max( mu_k*(lm_amp_factor),safety_factor); 
		   
		   
		    
		    
		    
		    if  norm(dx,2) <=  obj and obj <= 1.5*norm(dx,2) then 
			mu_k := obj_next;
		    fi:
		    
		    x_next := x_cur + dx;
		    obj := obj_next;
		    printf("k = %d, mu=%e, ||x_k - x_k+1|| = %e, ||f(x_k+1)||/||dx|| = %e, ||f(x_k)||=%e\n",i,mu_k, norm(dx,2), obj_next/norm(dx,2), obj_next);
		    
		    if norm(x_cur-x_next,2) < tol or obj < tol  then break fi:
		
		    x_cur := x_next;
		    
		    residuals := residuals_next; 
		    
		    
		    Z_next := Z_next_candidate; 
		    J:= fast_symbolic_hessian_full(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_next);
		    JTJ := LinearAlgebra:-Transpose(J).J;
		else 
		    
		    mu_k := mu_k * (lm_damp_factor);  
		    
		    if norm(dx,2)<tol or  obj < tol  then break fi:
		
		fi:
		
		
		
	    
		
	    od:
	    
	    current_pt := subs(Equate(Vector(Z),x_cur));
	    t_end := time():
	    printf("Computation took %f seconds\n",t_end-t_start);
	    printf("||DA||=%e \n", poly_norm_opt(subs(current_pt,DeltaA),t));
	    return [obj ,current_pt, t_end-t_start];
	    
end proc:
numeric_lm := proc(my_func,my_args,x,x_init,tol:=1e-16,max_iter := 100000,lm_param:= 1,lm_amp_factor:=1e-1,lm_damp_factor:=1e1)
      local i,
	    grad,
	    J,
	    JTJ,
	    rhs_k, 
	    lhs_k,
	    current_pt,
	    x_next,
	    Imat,
	    residuals,
	    obj,
	    j,
	    dx,
	    x_next_candidate,
	    x_cur,
	    mu_k,
	    safety_factor 
	    ;
	    
	    safety_factor:= 10^(-.5*Digits);
	    
	    J:= numeric_hessian22(my_func,my_args,x,x_init);
	    JTJ := LinearAlgebra:-Transpose(J).J;
	    printf("Computed Numerical Hessian...\n");
	    residuals := numeric_gradient2(my_func,my_args,x,x_init);
	    obj := norm(residuals,2);
	    
	    
	    printf("Computed numerical gradient...\n");
	    
	    Imat:=Matrix(LinearAlgebra:-Dimension(JTJ));
	    for j from 1 to LinearAlgebra:-RowDimension(Imat) do 
		Imat[j,j]:=1:
	    od:
	    
	  
	    
	    x_cur :=x_init;
	    mu_k := lm_param;
	    
	    printf("Beginning iteration\n");
	    for i from 1 to max_iter do
		rhs_k := JTJ + mu_k * Imat;
		lhs_k := -1*LinearAlgebra:-Transpose(J).residuals;
		dx := LinearAlgebra:-LinearSolve(rhs_k,lhs_k);
		
		
		x_next_candidate := x_cur + dx;
		
		residuals_next := numeric_gradient2(my_func,my_args,x,x_next_candidate);
		obj_next := norm(residuals_next,2);
		
		
	
		if obj_next <= obj  then 
		
		    mu_k :=max( mu_k*(lm_amp_factor),safety_factor); 
		   
		   
		    
		    
		    
		    if  norm(dx,2) <=  obj and obj <= (1.5)*norm(dx,2) then 
			mu_k := obj_next;
		    fi:
		    
		    x_next := x_cur + dx;
		    obj := obj_next;
		    printf("k = %d, mu=%e, ||x_k - x_k+1|| = %e, ||f(x_k+1)||/||dx|| = %e, ||f(x_k)||=%e\n",i,mu_k, norm(dx,2), obj_next/norm(dx,2), obj_next);
		    
		    if norm(x_cur-x_next,2) < tol or obj < tol then break fi:
		
		    x_cur := x_next;
		    
		    residuals := residuals_next; 
		    
		    J:= numeric_hessian22(my_func,my_args,x, x_cur);
		    JTJ := LinearAlgebra:-Transpose(J).J;
		
		
		else 
		    
		    mu_k := mu_k * (lm_damp_factor);  
		fi:
		
		
		
	    
		
	    od:
	    
	    current_pt := subs(Equate(Vector(x),x_cur));
	    return [obj , Equate(x,x_cur)];
	    
end proc:
numeric_gradient4 := proc(my_func,my_args,x,x_init,h:=sqrt(evalf(Digits))^(-1)) 
      local i,
	    Df,
	    x_h,
	    f1,
	    f2,
	    f3,
	    f4
	    ;
	    x_h := Vector(ArrayNumElems(x));
	    Df := Vector(ArrayNumElems(x));
	    
	    
	    
	    for i from 1 to ArrayNumElems(x) do 
		  x_h[i] := h; 
		  if i > 1 then x_h[i-1] :=0 fi: 
		  
		  f1:= my_func( op(my_args), Equate(x,x_init+2*x_h));
		  f2:= my_func( op(my_args), Equate(x,x_init+x_h));
		  f3:= my_func( op(my_args), Equate(x,x_init-x_h));
		  f4:= my_func( op(my_args), Equate(x,x_init-2*x_h));
		
		  Df[i] :=  1./(12*h) * (-1*f1 +  8*f2 - 8*f3 + f4);
	    od:
	    
	    return Df; 
	    
	    
end proc:
numeric_gradient2 := proc(my_func,my_args,x,x_init,h:=sqrt(evalf(Digits))^(-1)) 
      local i,
	    Df,
	    x_h,
	    f1,
	    f2
	    ;
	    x_h := Vector(ArrayNumElems(x));
	    Df := Vector(ArrayNumElems(x));
	    
	    
	    
	    for i from 1 to ArrayNumElems(x) do 
		  x_h[i] := h; 
		  if i > 1 then x_h[i-1] :=0 fi: 
		  
		  f1:= my_func( op(my_args), Equate(x,x_init+x_h));
		  f2:= my_func( op(my_args), Equate(x,x_init-1*x_h));
		
		  Df[i] :=  1./(2*h) * (f1-f2);
	    od:
	    
	    return Df; 
	    
	    
end proc:
numeric_jacobian4_adjoint := proc(A,t,A_init,DeltaA,DeltaA_vars,h:=sqrt(evalf(Digits))^(-1))
	local   i,
		J,
		d, n, degree_bound, 
		A_dim, A_h,
		DA_vars_dim, default_DA_pt,
		x1,x2,x3,x4, 
		G1,G2,G3,G4,G 
		;
		
		default_DA_pt := [seq(DeltaA_vars[i] = 0, i=1..nops(DeltaA_vars))]; 
		d:=matrix_degree(A,t);
		n:=LinearAlgebra:-RowDimension(A);
		size_bound := (n-1)*d+1; 
		J:=Matrix(0,0);
		for i from 1 to nops(DeltaA_vars) do 
		      
		      A_h :=  subs(default_DA_pt,subs(DeltaA_vars[i]=h,DeltaA));
		   
		      
		      x1 :=  A+2*A_h;
		      x2 :=  A+  A_h;
		      x3 :=  A-  A_h;
		      x4 :=  A-2*A_h;
		      
		      G1 := symbolic_numeric_adjoint(x1);
		      G2 := symbolic_numeric_adjoint(x2);
		      G3 := symbolic_numeric_adjoint(x3);
		      G4 := symbolic_numeric_adjoint(x4);
		
		
		      G:= convert(1/(12. * h) *(-1*G1+8*G2-8*G3+G4),Vector); 
		    
		    G:=convert(`<,>`( seq(Vector(size_bound,PolynomialTools:-CoefficientVector(G[j],t)),j=1..ArrayNumElems(G))),Vector);
		      if ArrayNumElems(J) = 0 then 
			  J := G;
		      else
			  J:= `<|>`(J,G);
		      fi:
			  
		od:
		
		
		return J;
end proc:
numeric_hessian44 := proc(my_func,my_args,x,x_init,h:=sqrt(evalf(Digits))^(-1)) 
	local   i,
		H,
		f_dim,
		x_dim,
		x_h,
		x1,x2,x3,x4, 
		G1,G2,G3,G4 
		;
		
		H:=Matrix();
		x_h := Vector(ArrayNumElems(x));
		for i from 1 to ArrayNumElems(x) do
		     if i > 1 then 
			x_h[i-1] := 0;
		     fi:
		     x_h[i]:=h;
		     
		     G1 := numeric_gradient4(my_func,(my_args),x,x_init+2*x_h);
		     G2 := numeric_gradient4(my_func,(my_args),x,x_init+x_h);
		     G3 := numeric_gradient4(my_func,(my_args),x,x_init-x_h);
		     G4 := numeric_gradient4(my_func,(my_args),x,x_init-2*x_h);
		     
		     G:= 1/(12.*h) * (-1*G1+8*G2-8*G3+G4);
		     
		     if ArrayNumElems(H) =0 then 
			H:=G;
		     else
			H:=`<|>`(H,G);
		     fi:
		     
		od:
		
		return H;
end proc:
numeric_hessian22 := proc(my_func,my_args,x,x_init,h:=sqrt(evalf(Digits))^(-1)) 
	local   i,
		H,
		f_dim,
		x_dim,
		x_h,
		G1,G2 
		;
		
		H:=Matrix();
		x_h := Vector(ArrayNumElems(x));
		for i from 1 to ArrayNumElems(x) do
		     if i > 1 then 
			x_h[i-1] := 0;
		     fi:
		     x_h[i]:=h;
		     
		     G1 := numeric_gradient2(my_func,(my_args),x,x_init+x_h,h);
		     G2 := numeric_gradient2(my_func,(my_args),x,x_init-x_h,h);
		     
		     G:= 1/(2.*h) * (G1-G2);
		     if ArrayNumElems(H) =0 then 
			H:=G;
		     else
			H:=`<|>`(H,G);
		     fi:
		     
		od:
		
		return H;
end proc:
numeric_hessian24 := proc(my_func,my_args,x,x_init,h:=sqrt(evalf(Digits))^(-1)) 
	local   i,
		H,
		f_dim,
		x_dim,
		x_h,
		G1,G2 
		;
		
		H:=Matrix();
		x_h := Vector(ArrayNumElems(x));
		for i from 1 to ArrayNumElems(x) do
		     if i > 1 then 
			x_h[i-1] := 0;
		     fi:
		     x_h[i]:=h;
		     
		     G1 := numeric_gradient4(my_func,(my_args),x,x_init+x_h,h);
		     G2 := numeric_gradient4(my_func,(my_args),x,x_init-x_h,h);
		     
		     G:= 1/(2.*h) * (G1-G2);
		     if ArrayNumElems(H) =0 then 
			H:=G;
		     else
			H:=`<|>`(H,G);
		     fi:
		     
		od:
		
		return H;
end proc:
extract_guess_information_original := proc(A,A_init,x:=t,max_gcd_degree:=2,method:="support",symbol:='a')
	local
	fstar_init,fstar_init_pt, fstar,fstar_vars,fstar_guesses,
	h_init,h_init_pt,h,h_vars,h_guesses,
	opt_index,
	Lambda,
	hfs_vec,J_fh,J_adj,J, 
	DA_vec,
	adjA,
	DeltaA,DeltaA_vars,DeltaA_init,DeltaA_init_pt,
	n,d,adjoint_degree_bound 
	;
	
	DeltaA,DeltaA_vars:= generate_perturbation(A,x,method,symbol); 
	DeltaA_init :=A_init-A;
	
	
	
	if check_entry_degrees(A,A_init,x) = false then 
	    WARNING("Initial guess has invalid support.");
	    
	    return [];
	fi:
	
	DeltaA_init_pt := [seq(op(Equate(Vector(1+max(degree(DeltaA(i),x), degree(DeltaA_init(i),x)),PolynomialTools:-CoefficientVector(DeltaA(i),x)),Vector(1+max(degree(DeltaA(i),x), degree(DeltaA_init(i),x) ),PolynomialTools:-CoefficientVector(DeltaA_init(i),x)))), i=1..ArrayNumElems(DeltaA))];
	
	
	d:= matrix_degree(A,x);
	n := min(LinearAlgebra:-Dimension(A)); 
	adjoint_degree_bound := (n-1)*d;
	
	
	adjA := convert(symbolic_numeric_adjoint(A_init),list); 
	
	
	
	
	
	
	adjA_degree := max(1,max(seq(degree(adjA[i],x),i=1..nops(adjA))));  
	
	
	
	
	
	
	
	
	
	h_guesses := [seq(generate_initial_agcd_degree(adjA,i,x), i=1..min(adjA_degree,max_gcd_degree) )];
	fstar_guesses := [seq(poly_least_squares_division_implicit(adjA, h_guesses[i]),i=1..nops(h_guesses))];
	
	opt_index:= min[index]([seq(fstar_guesses[i][1], i=1..nops(fstar_guesses))]);
	
	
	printf("Extraced co-factor and gcd Error:  %e, Degree:  %d\n", fstar_guesses[opt_index][1],opt_index);
	
	
	
	h_init := h_guesses[opt_index];
	fstar_init := fstar_guesses[opt_index][2];
	
	
	
	
	
	h := add(hh[i]*x^i,i=0..opt_index-1) + x^opt_index;
	
	h_vars := PolynomialTools:-CoefficientVector(h-x^opt_index,x);
	
	
	
	fstar := [seq(add(fs[j,i]*x^i,i=0..adjoint_degree_bound-opt_index),j=1..nops(fstar_init))]; 
	
	fstar_vars := `<,>`(seq(PolynomialTools:-CoefficientVector(fstar[i],x),i=1..nops(fstar)));
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
		
	h_init_pt := Equate(h_vars, Vector(PolynomialTools:-CoefficientVector(h_init,x)[1..opt_index]));
	
	fstar_init_pt:= Equate(fstar_vars,`<,>`(seq(Vector(adjoint_degree_bound-opt_index+1, PolynomialTools:-CoefficientVector(fstar_init[i],x)), i=1..nops(fstar_init) )));
	
	
	
	
	
	
	
	
	Lambda := Vector(n^2*(adjoint_degree_bound+1),i->'lambda'[i]); 
	
	X:= [op(DeltaA_vars), op(convert(`<,>`(h_vars,fstar_vars),list))];
	
	X_init := [op(DeltaA_init_pt), op(h_init_pt), op(fstar_init_pt)];
	
	
	
	
	
	
	
	hfs_vec := -1*convert(`<,>`( seq(PolynomialTools:-CoefficientVector(fstar[i]*h,x),i=1..nops(fstar))),Vector);
	
	
        
	J:=fast_symbolic_jacobian_constraints(A,x,DeltaA,DeltaA_vars,fstar,h,Lambda,X,X_init);
	
	
	
	DA_vec := convert(DeltaA,Vector);
	
	
	DA_vec := convert(DeltaA_vars,Vector);
	DA_vec := subs(DeltaA_init_pt,DA_vec);
	DA_vec := convert(`<,>`(DA_vec, Vector(ArrayNumElems(fstar_vars) +ArrayNumElems(h_vars))),Vector);
	
	
	
	
	
	my_soln := Optimization:-LSSolve([-2*DA_vec,LinearAlgebra:-Transpose(J)]);
	
	Lambda_init := Equate(Lambda,my_soln[2]);
	
	printf("Initial Lagrange multiplier residual: %e \n", my_soln[1]);
	
	  
	
	return [DeltaA,fstar,h], [X,X_init],[convert(Lambda,list),Lambda_init], [DeltaA_vars,fstar_vars,h_vars],[J];
	
	
	
	
	
end proc:
extract_guess_information := proc(A,A_init,x:=t,min_gcd_degree:=1,max_gcd_degree:=2,method:="support",symbol:='a')
	local
	fstar_init,fstar_init_pt, fstar,fstar_vars,fstar_guesses,
	h_init,h_init_pt,h,h_vars,h_guesses,
	opt_index,
	Lambda,
	hfs_vec,J_fh,J_adj,J, 
	DA_vec,
	adjA,
	DeltaA,DeltaA_vars,DeltaA_init,DeltaA_init_pt,
	n,d,adjoint_degree_bound 
	;
	
	DeltaA,DeltaA_vars:= generate_perturbation(A,x,method,symbol); 
	DeltaA_init :=A_init-A;
	
	
	
	if check_entry_degrees(A,A_init,x) = false then 
	    WARNING("Initial guess has invalid support.");
	    
	    return [];
	fi:
	
	if min_gcd_degree > max_gcd_degree then 
	    WARNING("Initial guess has invalid support.");
	    
	    return [];
	
	fi:
	
	
	DeltaA_init_pt := [seq(op(Equate(Vector(1+max(degree(DeltaA(i),x), degree(DeltaA_init(i),x)),PolynomialTools:-CoefficientVector(DeltaA(i),x)),Vector(1+max(degree(DeltaA(i),x), degree(DeltaA_init(i),x) ),PolynomialTools:-CoefficientVector(DeltaA_init(i),x)))), i=1..ArrayNumElems(DeltaA))];
	
	
	d:= matrix_degree(A,x);
	n := min(LinearAlgebra:-Dimension(A)); 
	adjoint_degree_bound := (n-1)*d;
	
	
	adjA := convert(symbolic_numeric_adjoint(A_init),list); 
	
	
	
	
	
	
	adjA_degree := max(1,max(seq(degree(adjA[i],x),i=1..nops(adjA))));  
	
	
	
	
	
	
	
	
	
	h_guesses := [seq(generate_initial_agcd_degree(adjA,i,x), i=1..min(adjA_degree,max_gcd_degree) )];
	fstar_guesses := [seq(poly_least_squares_division_implicit(adjA, h_guesses[i]),i=1..nops(h_guesses))];
	
	opt_index:= min[index]([seq(fstar_guesses[i][1], i=1..nops(fstar_guesses))]);
	
	if min_gcd_degree <= nops(h_guesses) and opt_index < min_gcd_degree then 
                opt_index := min_gcd_degree;
	fi:
	
	
	printf("Extraced co-factor and gcd Error:  %e, Degree:  %d\n", fstar_guesses[opt_index][1],opt_index);
	
	
	
	h_init := h_guesses[opt_index];
	fstar_init := fstar_guesses[opt_index][2];
	
	
	
	
	
	h := add(hh[i]*x^i,i=0..opt_index-1) + x^opt_index;
	
	h_vars := PolynomialTools:-CoefficientVector(h-x^opt_index,x);
	
	
	
	fstar := [seq(add(fs[j,i]*x^i,i=0..adjoint_degree_bound-opt_index),j=1..nops(fstar_init))]; 
	
	fstar_vars := `<,>`(seq(PolynomialTools:-CoefficientVector(fstar[i],x),i=1..nops(fstar)));
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
		
	h_init_pt := Equate(h_vars, Vector(PolynomialTools:-CoefficientVector(h_init,x)[1..opt_index]));
	
	fstar_init_pt:= Equate(fstar_vars,`<,>`(seq(Vector(adjoint_degree_bound-opt_index+1, PolynomialTools:-CoefficientVector(fstar_init[i],x)), i=1..nops(fstar_init) )));
	
	
	
	
	
	
	
	
	Lambda := Vector(n^2*(adjoint_degree_bound+1),i->'lambda'[i]); 
	
	X:= [op(DeltaA_vars), op(convert(`<,>`(h_vars,fstar_vars),list))];
	
	X_init := [op(DeltaA_init_pt), op(h_init_pt), op(fstar_init_pt)];
	
	
	
	
	
	
	
	hfs_vec := -1*convert(`<,>`( seq(PolynomialTools:-CoefficientVector(fstar[i]*h,x),i=1..nops(fstar))),Vector);
	
	
        
	J:=fast_symbolic_jacobian_constraints(A,x,DeltaA,DeltaA_vars,fstar,h,Lambda,X,X_init);
	
	
	
	DA_vec := convert(DeltaA,Vector);
	
	
	DA_vec := convert(DeltaA_vars,Vector);
	DA_vec := subs(DeltaA_init_pt,DA_vec);
	DA_vec := convert(`<,>`(DA_vec, Vector(ArrayNumElems(fstar_vars) +ArrayNumElems(h_vars))),Vector);
	
	
	
	
	
	my_soln := Optimization:-LSSolve([-2*DA_vec,LinearAlgebra:-Transpose(J)]);
	
	Lambda_init := Equate(Lambda,my_soln[2]);
	
	printf("Initial Lagrange multiplier residual: %e \n", my_soln[1]);
	
	  
	
	return [DeltaA,fstar,h], [X,X_init],[convert(Lambda,list),Lambda_init], [DeltaA_vars,fstar_vars,h_vars],[J];
	
	
	
	
	
end proc:
numeric_gradient2_turbo := proc()
end proc:
numeric_hessian22_turbo := proc()
end proc:
my_func := proc(A)
    return evalf(LinearAlgebra:-Adjoint(convert(A,rational)));
end proc:
numeric_lagrangian := proc(A,x,DeltaA,fstar,h,Lambda,Z_pt)
	  local DA_pt,
		fstar_pt,
		h_pt,
		adj_vec,
		coeff_vec,
		Lambda_pt,
		L,
		n,
		d
		;
    
		DA_pt := subs(Z_pt,DeltaA);
		fstar_pt := convert(subs(Z_pt,fstar),Vector);
		h_pt :=  subs(Z_pt,h);
		
		Lambda_pt := convert(subs(Z_pt,Lambda),Vector);
		
		
		adj_vec := convert(symbolic_numeric_adjoint(A+DA_pt),Vector);
		
		coeff_vec := vectorize_adjoint(A,adj_vec-fstar_pt * h_pt,x); 
		
		L:= poly_norm_opt(DA_pt,x)^2 + (LinearAlgebra:-Transpose(Lambda_pt).coeff_vec)[1]; 
		
		return L;
		
end proc:
symbolic_numeric_adjoint := proc (A)
	local adj_A
	      ;
	      
	      return LinearAlgebra:-Adjoint(evalf(A));
end proc:
vectorize_adjoint := proc(A,Adj,t)
      local 
	    adj_vec,
	    n,
	    d,
	    size_bound
	    ;
	    
	    
	    n := LinearAlgebra:-RowDimension(A); 
	    d := matrix_degree(A,t);
	    size_bound := (n-1)*d+1;
	    
	    
	    adj_vec:= `<,>` (seq((  Vector( PolynomialTools:-CoefficientVector(Adj(i),t),size_bound)), i=1..ArrayNumElems(Adj)));
	   
	    
	    return convert(adj_vec,Vector); 
end proc:
check_entry_degrees := proc(A,A_init,x)
    local i;
	  for i from 1 to ArrayNumElems(A) do 
	      if degree (A(i),x) < degree (A_init(i),x) then return false fi:
	  od:
	  return true;
end proc:
fast_symbolic_gradient := proc(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init)
	  
	  local adjA_vec,
		hfs_vec,
		
		J,
		G
		;
	  adjA_vec := vectorize_adjoint(A,symbolic_numeric_adjoint(A+subs(Z_init,DeltaA)),t);
	 
	  hfs_vec := subs(Z_init,compute_hfs_vec(fstar,h,t));
	  J:= fast_symbolic_jacobian_constraints(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init);
	  
	  G:= 2*Vector(convert(subs(Z_init,X[1..nops(DeltaA_vars)]),Vector),nops(X));
	  G:= G+ LinearAlgebra:-Transpose(J) . convert(subs(Z_init,Lambda),Vector);
	  
	  G:=convert(`<,>`(G,adjA_vec-hfs_vec),Vector);
	  return G;
end proc:
compute_hfs_vec := proc(fstar,h,t)
    return     convert(`<,>`( seq(PolynomialTools:-CoefficientVector(fstar[i]*h,t),i=1..nops(fstar))),Vector);
end proc:
fast_symbolic_hessian_full:= proc(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init)
	local H_A,
	      H_fh,
	      H,
	      J,
	      Lambda_pt,
	      fh_vars,
	      hfs_vec
	      ;
	      
	      fh_vars := X[nops(DeltaA_vars)+1..nops(X)];
	      Lambda_pt := subs(Z_init,convert(Lambda,Vector[row]));
	      hfs_vec := -1*compute_hfs_vec(fstar,h,t);
	      
	      H_A := fast_symbolic_hessianXX(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init);
	      
	      
	      
	      H_fh := VectorCalculus:-Hessian(Lambda_pt.hfs_vec,fh_vars);
	      
	      J := fast_symbolic_jacobian_constraints(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init);
	      
	      
	      H:= LinearAlgebra:-DiagonalMatrix([H_A,H_fh]);
	      H:= `<,>`(H,J);
	      H:=`<|>`(H,`<,>`(LinearAlgebra:-Transpose(J),Matrix(LinearAlgebra:-RowDimension(J))));
	      return H;
end proc:
fast_symbolic_hessianXX := proc(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init)
     local i, j,k,ell,
	    hfs_vec,
	    fh,
	    fh_vars,
	    n,
	    Adj,
	    J,J_temp, 
	    active_vars, active_vars_ij,
	    active_pt, 
	    A_copy,
	    d, 
	    H, 
	    Lambda_vec,
	    adjA_vec, 
	    size_bound
	    ;
	    
	    d:=matrix_degree(A,t);
	    n := min(LinearAlgebra:-Dimension(A));
	    size_bound := (n-1)*d+1;
	    fh := fstar * h;
	    fh_vars := X[nops(DeltaA_vars)+1..nops(X)];
	    J:=Matrix(0,0):
	    
	    hfs_vec := -1*convert(`<,>`( seq(PolynomialTools:-CoefficientVector(fstar[i]*h,t),i=1..nops(fstar))),Vector);
	    
	    
	    
	    
	    
	    for i from 1 to LinearAlgebra:-RowDimension(DeltaA) do 
		for j from 1 to LinearAlgebra:-ColumnDimension(DeltaA) do
		    active_vars_ij := convert(PolynomialTools:-CoefficientVector(DeltaA[i,j],t),set) minus {0};
		    adjA_vec:=Vector(n^2*size_bound); 
	   
		    for k from i+1 to LinearAlgebra:-RowDimension(DeltaA) do
			for ell from 1 to LinearAlgebra:-ColumnDimension(DeltaA) do 
			    
			    
			    if ell <> j then
			    
				active_vars := convert(PolynomialTools:-CoefficientVector(DeltaA[k,ell],t),set) minus {0};
				active_vars := active_vars_ij union active_vars;
				
			    
				
				
				inactive_pt := convert(convert(DeltaA_vars,set) minus active_vars,Vector);   
				inactive_pt := Equate(inactive_pt, subs(Z_init,inactive_pt));
				
				
				active_pt := convert(active_vars,Vector);
				active_pt :=  Equate(active_pt, subs(Z_init,active_pt));
				
				
				active_vars := convert(active_vars,list);
				
			    
				A_copy := A+subs(inactive_pt,DeltaA); 
			    
				
				
				adjA_vec:=adjA_vec+ vectorize_adjoint(A,symbolic_numeric_adjoint(A_copy),t)-hfs_vec;
			    fi:
			od:
		    od:
		    
		    active_vars_ij := convert(active_vars_ij,list);
		    if ArrayNumElems(J) =0 then 
			J:= VectorCalculus:-Jacobian(adjA_vec,active_vars_ij);
		    else 
			J:= `<|>`(J,VectorCalculus:-Jacobian(adjA_vec,active_vars_ij));
			
		    fi:
		    
		od:
	   od:
	    
	 
	    
	    Lambda_vec := convert(subs(Z_init,Lambda),Vector[row]): 
	    
	    H:=VectorCalculus:-Jacobian( Lambda_vec.J, DeltaA_vars);
	   
	   H:=H+LinearAlgebra:-Transpose(H)+2*LinearAlgebra:-IdentityMatrix(LinearAlgebra:-Dimension(H)); 
	   
	    return subs(Z_init,H);
end proc:
fast_symbolic_jacobian_constraints := proc(A,t,DeltaA,DeltaA_vars,fstar,h,Lambda,X,Z_init)
      local i,
	    j,
	    hfs_vec,
	    fh,
	    fh_vars,
	   
	    n,
	    Adj,
	    J, 
	    active_vars, 
	    active_pt, 
	    A_copy,
	    d,
	    adjA_vec, 
	    size_bound
	    ;
	    
	    d:=matrix_degree(A,t);
	    n := min(LinearAlgebra:-Dimension(A));
	    size_bound := (n-1)*d+1;
	    fh := fstar * h;
	    fh_vars := X[nops(DeltaA_vars)+1..nops(X)];
	    J:=Matrix(0,0):
	    
	    hfs_vec := -1*convert(`<,>`( seq(PolynomialTools:-CoefficientVector(fstar[i]*h,t),i=1..nops(fstar))),Vector);
	    adjA_vec := Vector(n^2 * size_bound);
	    
	    for i from 1 to LinearAlgebra:-RowDimension(DeltaA) do 
		for j from 1 to LinearAlgebra:-ColumnDimension(DeltaA) do
		    active_vars := convert(PolynomialTools:-CoefficientVector(DeltaA[i,j],t),set) minus {0};
		    
		    
		    
		    
		    
		    inactive_pt := convert(convert(DeltaA_vars,set) minus active_vars,Vector);   
		    inactive_pt := Equate(inactive_pt, subs(Z_init,inactive_pt));
		    
		    
		    active_pt := convert(active_vars,Vector);
		    active_pt :=  Equate(active_pt, subs(Z_init,active_pt));
		    
		    
		    active_vars := convert(active_vars,list);
		    
		 
		    A_copy := A+subs(inactive_pt,DeltaA); 
		 
		    
		    
		    adjA_vec:= vectorize_adjoint(A,symbolic_numeric_adjoint(A_copy),t)-hfs_vec;
		    
		    
		    
		    if ArrayNumElems(J) =0 then 
			J:= VectorCalculus:-Jacobian(adjA_vec,active_vars);
		    else 
			J:= `<|>`(J,VectorCalculus:-Jacobian(adjA_vec,active_vars));
			
		    fi:
		od:
	   od:
	    
	    J:= `<|>`(J,VectorCalculus:-Jacobian(hfs_vec,fh_vars)); 
	    
	    return subs(Z_init,J);
end proc:
linearize_mccoy:=proc(A,t,DeltaA,DA_vars,s_rank)
        local A_lin1,A_lin0, 
              DA_lin1,DA_lin0, 
              omega, r_omega,i_omega, 
              m,n, 
              A_0,A_1,DA_0,DA_1,
              K, K_vars, 
              
              real_rank_constraint,
              imaginary_rank_constraint,
              rank_constraint,
              real_singularity_constraint,
              imaginary_singularity_constraint,
              singularity_constraint,
              constraints, 
              Lambda,
              omega_vars,
              DA_lin,
              A_lin,
              L,
              vars,
              residuals,
              X
              ;
              A_lin1,A_lin0 := LinearAlgebra:-CompanionMatrix(A,t);
            DA_lin1,DA_lin0 := LinearAlgebra:-CompanionMatrix(A+DeltaA,t);
              A_lin := A_lin1-t*A_lin0;
              DA_lin := simplify(DA_lin1-t*DA_lin0-A_lin); 
              
              
            omega := r_omega + I*i_omega;
            m,n := LinearAlgebra:-Dimension(A_lin):
            
            K:=Matrix( m,s_rank, (i,j)->(rk[i,j] + I*ik[i,j]) ):
            K_vars := convert(Re(K),list) assuming real: 
            K_vars := [op(K_vars), op(convert(Im(K),list)) ] assuming real:
            real_rank_constraint := convert(Re(evalc(LinearAlgebra:-HermitianTranspose(K).K-LinearAlgebra:-IdentityMatrix(s_rank))),Vector) assuming real:
            imaginary_rank_constraint := convert(Im(evalc(LinearAlgebra:-HermitianTranspose(K).K-LinearAlgebra:-IdentityMatrix(s_rank))),Vector) assuming real:
            rank_constraint := `<,>`(real_rank_constraint,imaginary_rank_constraint):
            A_0 := subs(t=0,A_lin):
            A_1 := (A_lin-A_0)/t: 
            DA_0 := subs(t=0,DA_lin):
            DA_1 := (DA_lin-DA_0)/t: 
            real_singularity_constraint := convert(Re( (A_0 + DA_0).K + evalc(omega * (A_1+DA_1).K)),Vector) assuming real:
            imaginary_singularity_constraint :=  convert(Im( (A_0 + DA_0).K + evalc(omega * (A_1+DA_1).K)),Vector) assuming real:
            singularity_constraint:= `<,>`(real_singularity_constraint,imaginary_singularity_constraint):
            constraints := `<,>`((singularity_constraint),(rank_constraint)):
            Lambda := Vector[row](ArrayNumElems(constraints), i->lambda[i]):
            omega_vars := [r_omega,i_omega]:
            
            X:= [op(DA_vars),op(omega_vars), op(K_vars)];
            vars := [op(X), op(convert(Lambda,list))];;
                
            L:= evalf(poly_norm_opt(DeltaA,t)^2 + subs([],(Lambda.constraints))[1]):
            residuals := VectorCalculus:-Gradient(L,vars):
            return [residuals,[A_lin,DA_lin,K,omega], [X,Lambda]];
end proc:
linearize_mccoy_x_init := proc(A_lin,t,s_rank,DA_lin,omega_init,K,omega,A,A_init,DeltaA)
            
            local omega_init_pt,
            my_U,my_Vt, 
            K_init,
            approx_kernel,
            DA_init,
            DeltaA_init
            
            ;
            
            omega_init_pt := [Re(omega)=Re(omega_init),Im(omega) = Im(omega_init)] assuming real;
            my_U,my_Vt:= LinearAlgebra:-SingularValues(subs(t=omega_init,A_lin),output=['U','Vt']);
            
       
            approx_kernel := LinearAlgebra:-HermitianTranspose(my_Vt[LinearAlgebra:-RowDimension(my_Vt)-s_rank+1..LinearAlgebra:-RowDimension(my_Vt)]);
            
            K_init :=[op( Equate(Re(K), Re(approx_kernel))),op( Equate(Im(K), Im(approx_kernel)))]  assuming real;
            
            
            DeltaA_init :=A_init-A;
            DA_init := [seq(op(Equate(Vector(1+mat(degree(DeltaA(i),t), degree(DeltaA_init(i),t)),PolynomialTools:-CoefficientVector(DeltaA(i),t)),Vector(1+max(degree(DeltaA(i),t), degree(DeltaA_init(i),t) ),PolynomialTools:-CoefficientVector(DeltaA_init(i),t)))), i=1..ArrayNumElems(DeltaA))];
            
            return [op(DA_init), op(K_init), op(omega_init_pt)];
end proc:
linearize_mccoy_lambda_init := proc(residuals,X_init,Lambda)
        local   AA,bb, 
                Lambda_init
                ;
                
                AA,bb := LinearAlgebra:-GenerateMatrix(Equate(Vector(subs(X_init,residuals)),Vector(ArrayNumElems(residuals))),convert(Lambda,list));
                
                
                print(LinearAlgebra:-LeastSquares(AA,bb));
               Lambda_init :=  Equate(convert(Lambda,Vector), LinearAlgebra:-LeastSquares(AA,bb));
        return Lambda_init;
end proc:
lower_mccoy := proc(A,t,A_init,omega_init,s_rank)
            local   my_args,
                    my_residuals,
                    A_lin,DA_lin,K,omega,
                    X,Lambda,X_init,
                    Lambda_init,
                    Z_init,
                    Z;
                    
                    my_args := linearize_mccoy(A,t,DeltaA,DA_vars,s_rank):
                    my_residuals := my_args[1]: 
                    A_lin,DA_lin,K,omega := op(my_args[2]):
                    X,Lambda := op(my_args[3]):
                    X_init := linearize_mccoy_x_init(A_lin,t,s_rank,DA_lin,omega_init,K,omega,A,A_init,DeltaA):
                    lambda_init := linearize_mccoy_lambda_init(my_residuals,X_init,Lambda);
                    Z_init := [op(X_init), op(lambda_init)];
                    Z:= [op(X),op(convert(Lambda,list))];
                    
                    return [my_residuals,Z,Z_init,omega];
end proc:
